"""
compare simulation outputs to verify model performance
"""

from __future__ import annotations
from pathlib import Path
import numpy as np
import logging
import sys
import argparse
from datetime import datetime
import typing as T

from . import read
from . import find

try:
    from .plot.diff import plotdiff
except ImportError:
    plotdiff = None

TOL = {
    "rtol": 1e-5,
    "rtolN": 1e-5,
    "rtolT": 1e-5,
    "rtolJ": 1e-5,
    "rtolV": 1e-5,
    "atol": 1e-8,
    "atolN": 1e9,
    "atolT": 100,
    "atolJ": 1e-7,
    "atolV": 50,
}


def cli():

    p = argparse.ArgumentParser(description="Compare simulation file outputs and inputs")
    p.add_argument("outdir", help="directory to compare")
    p.add_argument("refdir", help="reference directory")
    p.add_argument("-only", help="only check in or out", choices=["in", "out"])
    p.add_argument(
        "-file_format",
        help="specify file format to read from output dir",
        choices=["h5", "nc", "raw"],
    )
    P = p.parse_args()

    errs = compare_all(P.outdir, refdir=P.refdir, tol=TOL, file_format=P.file_format, only=P.only)

    if errs:
        for e, v in errs.items():
            print(f"{e} has {v} errors", file=sys.stderr)

        raise SystemExit(f"FAIL: compare {P.outdir}")

    print(f"OK: Gemini comparison {P.outdir} {P.refdir}")


def compare_all(
    outdir: Path,
    refdir: Path,
    tol: dict[str, float] = TOL,
    *,
    plot: bool = True,
    file_format: str = None,
    only: str = None,
) -> dict[str, int]:
    """
    compare two directories across time steps
    """
    outdir = Path(outdir).expanduser()
    refdir = Path(refdir).expanduser()

    if outdir.samefile(refdir):
        raise OSError(f"reference and output are the same directory: {outdir}")

    # %% fail hard if grid doesn't match, because otherwise data is non-sensical
    if compare_grid(outdir, refdir, tol, file_format=file_format) != 0:
        raise ValueError("grid values do not match {outdir}  {refdir}")

    errs = {}
    if not only or only == "out":
        e = compare_output(outdir, refdir, tol, file_format=file_format, plot=plot)
        if e:
            errs["out"] = e

    if not only or only == "in":
        e = compare_input(outdir, refdir, tol, file_format=file_format, plot=plot)
        if e:
            errs["in"] = e

    return errs


def compare_grid(
    outdir: Path, refdir: Path, tol: dict[str, float] = TOL, *, file_format: str = None
) -> int:

    ref = read.grid(refdir)
    new = read.grid(outdir, file_format=file_format)

    if not ref:
        raise FileNotFoundError(f"No simulation grid in {refdir}")
    if not new:
        raise FileNotFoundError(f"No simulation grid in {outdir}")

    errs = 0

    for k in ref.keys():
        if not isinstance(ref[k], np.ndarray):
            continue

        assert (
            ref[k].shape == new[k].shape
        ), f"{k}: ref shape {ref[k].shape} != data shape {new[k].shape}"
        if not np.allclose(ref[k], new[k], rtol=tol["rtol"], atol=tol["atol"]):
            errs += 1
            logging.error(f"{k}  {err_pct(ref[k], new[k]):.1f} %")

    return errs


def compare_input(
    outdir: Path,
    refdir: Path,
    tol: dict[str, float] = TOL,
    *,
    file_format: str = None,
    plot: bool = True,
) -> int:

    names = ("ns", "Ts", "vs1")

    ref_params = read.config(refdir)
    if not ref_params:
        raise FileNotFoundError(f"{refdir} does not appear to contain config.nml")
    ref_indir = refdir / ref_params["indat_file"].parts[-2]
    ref = read.data(ref_indir / ref_params["indat_file"].name, var=names)

    new_params = read.config(outdir)
    if not new_params:
        raise FileNotFoundError(f"{outdir} does not appear to contain config.nml")
    if len(new_params["time"]) <= 1:
        raise ValueError(
            f"{outdir} simulation did not run long enough, must run for more than one time step"
        )
    new_indir = outdir / new_params["indat_file"].parts[-2]
    new = read.data(new_indir / new_params["indat_file"].name, var=names)

    errs = 0
    # %% initial conditions
    itols = ("N", "T", "V")

    for k, j in zip(names, itols):
        b = ref[k]
        a = new[k]

        assert a.shape == b.shape, f"{k}: ref shape {b.shape} != data shape {a.shape}"

        if not np.allclose(a, b, 0.1 * tol[f"rtol{j}"], 0.1 * tol[f"atol{j}"]):
            errs += 1
            logging.error(f"{k}  {err_pct(a, b):.1f} %")

            if plot and plotdiff is not None:
                plotdiff(a, b, k, ref_params["time"][0], outdir, refdir)

    if "precdir" in new_params:
        prec_errs = compare_precip(
            ref_params["time"],
            new_indir / new_params["precdir"].name,
            ref_indir / ref_params["precdir"].name,
            tol,
            plot=plot,
            file_format=file_format,
        )
        errs += prec_errs

    if "E0dir" in new_params:
        efield_errs = compare_Efield(
            ref_params["time"],
            new_indir / new_params["E0dir"].name,
            ref_indir / ref_params["E0dir"].name,
            tol,
            plot=plot,
            file_format=file_format,
        )
        errs += efield_errs

    return errs


def err_pct(a: np.ndarray, b: np.ndarray) -> float:
    """ compute maximum error percent """

    return (abs(a - b).max() / abs(b).max()).item() * 100


def compare_precip(
    times: list[datetime],
    newdir: Path,
    refdir: Path,
    tol: dict[str, float] = TOL,
    *,
    plot: bool = False,
    file_format: str = None,
) -> int:

    prec_errs = 0

    # often we reuse precipitation inputs without copying over files
    for t in times:
        ref = read.precip(find.frame(refdir, t))
        new = read.precip(find.frame(newdir, t), file_format=file_format)

        for k in ref.keys():
            b = np.atleast_1d(ref[k])
            a = np.atleast_1d(new[k])

            assert a.shape == b.shape, f"{k}: ref shape {b.shape} != data shape {a.shape}"

            if not np.allclose(a, b, tol["rtol"], tol["atol"]):
                prec_errs += 1
                logging.error(f"{k} {t}  {err_pct(a, b):.1f} %")
                if plot and plotdiff is not None:
                    plotdiff(a, b, k, t, newdir.parent, refdir.parent)
            if prec_errs == 0:
                logging.info(f"OK: {k}  {newdir}")

    return prec_errs


def compare_Efield(
    times: list[datetime],
    newdir: Path,
    refdir: Path,
    tol: dict[str, float] = TOL,
    *,
    plot: bool = False,
    file_format: str = None,
) -> int:

    efield_errs = 0
    # often we reuse Efield inputs without copying over files
    for t in times:
        ref = read.Efield(find.frame(refdir, t))
        new = read.Efield(find.frame(newdir, t), file_format=file_format)
        for k in ("Exit", "Eyit", "Vminx1it", "Vmaxx1it"):
            b = ref[k]
            a = new[k]

            assert a.shape == b.shape, f"{k}: ref shape {b.shape} != data shape {a.shape}"

            if not np.allclose(a, b, tol["rtol"], tol["atol"]):
                efield_errs += 1
                logging.error(f"{k} {t}  {err_pct(a, b):.1f} %")
                if plot and plotdiff is not None:
                    plotdiff(a, b, k, t, newdir.parent, refdir.parent)

    if efield_errs == 0:
        logging.info(f"OK: Efield {newdir}")

    return efield_errs


def compare_output(
    outdir: Path,
    refdir: Path,
    tol: dict[str, float],
    *,
    file_format: str = None,
    plot: bool = True,
) -> int:
    """compare simulation outputs"""

    ref: dict[str, T.Any] = {}
    errs = 0
    a: np.ndarray = None

    params = read.config(outdir)
    if not params:
        raise FileNotFoundError(f"{outdir} does not appear to contain config.nml")
    if len(params["time"]) <= 1:
        raise ValueError(
            f"{outdir} simulation did not run long enough, must run for more than one time step"
        )

    for i, t in enumerate(params["time"]):
        st = f"UTsec {t}"
        A = read.frame(outdir, t, file_format=file_format)
        if not A:
            raise FileNotFoundError(f"{outdir} does not appear to contain data at {t}")
        B = read.frame(refdir, t)
        # don't specify file_format for reference,
        # so that one reference file format can check multiple "out" format

        names = ["ne", "v1", "v2", "v3", "Ti", "Te", "J1", "J2", "J3"]
        itols = ["N", "V", "V", "V", "T", "T", "J", "J", "J"]

        for k, j in zip(names, itols):
            a = A[k]
            b = B[k]

            assert a.shape == b.shape, f"{k} time {i} {t}: shape: ref {b.shape} != data {a.shape}"

            if not np.allclose(a, b, tol[f"rtol{j}"], tol[f"atol{j}"]):
                errs += 1
                logging.error(f"{k} {st}   {err_pct(a, b):.1f}")
                if plot and plotdiff is not None:
                    plotdiff(a, b, k, t, outdir, refdir)
        # %% assert time steps have unique output (earth always rotating...)
        if i > 1:
            names = ["ne", "v1", "v2", "v3"]
            itols = ["N", "V", "V", "V"]
            for k, j in zip(names, itols):
                if np.allclose(ref[k], a, 0.0001 * tol[f"rtol{j}"], 0.0001 * tol[f"atol{j}"]):
                    errs += 1
                    logging.error(f"{k} {st} too similar to prior step")

        if i == 3:
            for k in ("Ti", "Te"):
                if np.allclose(ref[k], A[k], 0.01 * tol["rtolT"], 0.1 * tol["atolT"]):
                    errs += 1
                    logging.error(f"{k} {st} too similar to prior step")

        if i == 2:
            for k in ("J1", "J2", "J3"):
                if np.allclose(ref[k], a, 0.01 * tol["rtolJ"], 0.1 * tol["atolJ"]):
                    errs += 1
                    logging.error(f"{k} {st} too similar to prior step")

        ref.update(A)

    return errs


if __name__ == "__main__":
    cli()



"""
raw binary file I/O.
Raw files are deprecated and do not contain most features of Gemini
"""

from __future__ import annotations
import xarray
from pathlib import Path
import typing as T
import numpy as np
import logging
import struct
from datetime import datetime, timedelta

from .. import find
from .. import WAVELEN, LSP


def simsize(path: Path) -> tuple[int, ...]:
    """
    get simulation size

    Parameters
    ----------
    fn: pathlib.Path
        filepath to simsize.dat

    Returns
    -------
    size: tuple of int, int, int
        3 integers telling simulation grid size
    """

    path = find.simsize(path, ".dat")

    fsize = path.stat().st_size
    if fsize == 12:
        lxs = struct.unpack("III", path.open("rb").read(12))
    elif fsize == 8:
        lxs = struct.unpack("II", path.open("rb").read(8))
    else:
        raise ValueError(f"{path} is not expected 8 bytes (2-D) or 12 bytes (3-D) long")

    return lxs


def grid(fn: Path, shape: bool = False) -> dict[str, T.Any]:
    """
    get simulation grid

    Parameters
    ----------
    fn: pathlib.Path
        filepath to simgrid

    Returns
    -------
    grid: dict
        grid parameters
    """

    if shape:
        raise NotImplementedError("grid shape for raw would be straightforward.")

    lxs = simsize(fn.parent)
    if len(lxs) == 2:
        return grid2(fn, lxs)
    elif len(lxs) == 3:
        return grid3(fn, lxs)
    else:
        raise ValueError("lxs must be 2-D or 3-D")


def grid2(fn: Path, lxs: tuple[int, ...] | list[int]) -> dict[str, np.ndarray]:
    """ for Efield """
    if not fn.is_file():
        raise FileNotFoundError(fn)

    grid: dict[str, T.Any] = {"lx": lxs}
    with fn.open("r") as f:
        grid["mlon"] = np.fromfile(f, np.float64, lxs[0])
        grid["mlat"] = np.fromfile(f, np.float64, lxs[1])

    return grid


def grid3(fn: Path, lxs: tuple[int, ...] | list[int]) -> dict[str, np.ndarray]:

    lgridghost = (lxs[0] + 4) * (lxs[1] + 4) * (lxs[2] + 4)
    gridsizeghost = [lxs[0] + 4, lxs[1] + 4, lxs[2] + 4]

    grid: dict[str, T.Any] = {"lx": lxs}

    if not fn.is_file():
        logging.error(f"{fn} grid file is not present. Will try to load rest of data.")
        return grid

    read = np.fromfile

    with fn.open("r") as f:
        for i in (1, 2, 3):
            grid[f"x{i}"] = read(f, np.float64, lxs[i - 1] + 4)
            grid[f"x{i}i"] = read(f, np.float64, lxs[i - 1] + 1)
            grid[f"dx{i}b"] = read(f, np.float64, lxs[i - 1] + 3)
            grid[f"dx{i}h"] = read(f, np.float64, lxs[i - 1])
        for i in (1, 2, 3):
            grid[f"h{i}"] = read(f, np.float64, lgridghost).reshape(gridsizeghost)
        L = [lxs[0] + 1, lxs[1], lxs[2]]
        for i in (1, 2, 3):
            grid[f"h{i}x1i"] = read(f, np.float64, np.prod(L)).reshape(L)
        L = [lxs[0], lxs[1] + 1, lxs[2]]
        for i in (1, 2, 3):
            grid[f"h{i}x2i"] = read(f, np.float64, np.prod(L)).reshape(L)
        L = [lxs[0], lxs[1], lxs[2] + 1]
        for i in (1, 2, 3):
            grid[f"h{i}x3i"] = read(f, np.float64, np.prod(L)).reshape(L)
        for i in (1, 2, 3):
            grid[f"gx{i}"] = read(f, np.float64, np.prod(lxs)).reshape(lxs)
        for k in ("alt", "glat", "glon", "Bmag"):
            grid[k] = read(f, np.float64, np.prod(lxs)).reshape(lxs)
        grid["Bincl"] = read(f, np.float64, lxs[1] * lxs[2]).reshape(lxs[1:])
        grid["nullpts"] = read(f, np.float64, np.prod(lxs)).reshape(lxs)
        if f.tell() == fn.stat().st_size:  # not EOF
            return grid

        L = [lxs[0], lxs[1], lxs[2], 3]
        for i in (1, 2, 3):
            grid[f"e{i}"] = read(f, np.float64, np.prod(L)).reshape(L)
        for k in ("er", "etheta", "ephi"):
            grid[k] = read(f, np.float64, np.prod(L)).reshape(L)
        for k in ("r", "theta", "phi"):
            grid[k] = read(f, np.float64, np.prod(lxs)).reshape(lxs)
        if f.tell() == fn.stat().st_size:  # not EOF
            return grid

        for k in ("x", "y", "z"):
            grid[k] = read(f, np.float64, np.prod(lxs)).reshape(lxs)

    return grid


def Efield(file: Path) -> xarray.Dataset:
    """
    load Efield_inputs files that contain input electric field in V/m
    """

    lxs = simsize(file.parent)

    assert lxs[0] > 0, "must have strictly positive number of longitude cells"
    assert lxs[1] > 0, "must have strictly positive number of latitude cells"

    m = grid2(file.parent / "simgrid.dat", lxs)

    assert (
        (m["mlat"] >= -90) & (m["mlat"] <= 90)
    ).all(), f"impossible latitude, was file read correctly? {file}"

    dat = xarray.Dataset(coords=m)

    with file.open("r") as f:
        """
        NOTE:
        this is mistakenly a float from Matlab
        to keep compatibility with old files, we left it as real64.
        New work should be using HDF5 instead of raw in any case.
        """
        dat["flagdirich"] = int(np.fromfile(f, np.float64, 1))
        for p in ("Exit", "Eyit", "Vminx1it", "Vmaxx1it"):
            dat[p] = (("x2", "x3"), read2D(f, lxs))
        for p in ("Vminx2ist", "Vmaxx2ist"):
            dat[p] = (("x2",), np.fromfile(f, np.float64, lxs[1]))
        for p in ("Vminx3ist", "Vmaxx3ist"):
            dat[p] = (("x3",), np.fromfile(f, np.float64, lxs[0]))
        filesize = file.stat().st_size
        if f.tell() != filesize:
            logging.error(f"{file} size {filesize} != file read position {f.tell()}")

    return dat


def frame3d_curv(file: Path) -> xarray.Dataset:
    """
    curvilinear

    Parameters
    ----------

    file: pathlib.Path
        filename to read
    """

    xg = grid(file.parent)
    dat = xarray.Dataset(coords={"x1": xg["x1"][2:-2], "x2": xg["x2"][2:-2], "x3": xg["x3"][2:-2]})
    lxs = simsize(file.parent)

    with file.open("r") as f:
        time(f)

        ns = read4D(f, LSP, lxs)
        dat["ne"] = (("x1", "x2", "x3"), ns[:, :, :, LSP - 1])

        vs1 = read4D(f, LSP, lxs)
        dat["v1"] = (
            ("x1", "x2", "x3"),
            (ns[:, :, :, :6] * vs1[:, :, :, :6]).sum(axis=3) / dat["ne"],
        )

        Ts = read4D(f, LSP, lxs)
        dat["Ti"] = (
            ("x1", "x2", "x3"),
            (ns[:, :, :, :6] * Ts[:, :, :, :6]).sum(axis=3) / dat["ne"],
        )
        dat["Te"] = (("x1", "x2", "x3"), Ts[:, :, :, LSP - 1].squeeze())

        for p in ("J1", "J2", "J3", "v2", "v3"):
            dat[p] = (("x1", "x2", "x3"), read3D(f, lxs))

        dat["Phitop"] = (("x2", "x3"), read2D(f, lxs))

    return dat


def frame3d_curvavg(file: Path) -> xarray.Dataset:
    """

    Parameters
    ----------
    file: pathlib.Path
        filename of this timestep of simulation output
    """

    lxs = simsize(file.parent)
    xg = grid(file.parent)
    dat = xarray.Dataset(coords={"x1": xg["x1"][2:-2], "x2": xg["x2"][2:-2], "x3": xg["x3"][2:-2]})

    with file.open("r") as f:
        time(f)

        for p in ("ne", "v1", "Ti", "Te", "J1", "J2", "J3", "v2", "v3"):
            dat[p] = (("x1", "x2", "x3"), read3D(f, lxs))

        dat["Phitop"] = (("x2", "x3"), read2D(f, lxs))

    return dat


def frame3d_curvne(file: Path) -> xarray.Dataset:

    lxs = simsize(file.parent)
    xg = grid(file.parent)
    dat = xarray.Dataset(coords={"x1": xg["x1"][2:-2], "x2": xg["x2"][2:-2], "x3": xg["x3"][2:-2]})

    with file.open("r") as f:
        time(f)

        dat["ne"] = (("x1", "x2", "x3"), read3D(f, lxs))

    return dat


def read4D(f, lsp: int, lxs: tuple[int, ...] | list[int]) -> np.ndarray:
    """
    read 4D array from raw file
    """
    if not len(lxs) == 3:
        raise ValueError(f"lxs must have 3 elements, you have lxs={lxs}")

    return np.fromfile(f, np.float64, np.prod(lxs) * lsp).reshape((*lxs, lsp), order="F")


def read3D(f, lxs: tuple[int, ...] | list[int]) -> np.ndarray:
    """
    read 3D array from raw file
    """
    if not len(lxs) == 3:
        raise ValueError(f"lxs must have 3 elements, you have lxs={lxs}")

    return np.fromfile(f, np.float64, np.prod(lxs)).reshape(*lxs, order="F")


def read2D(f, lxs: tuple[int, ...] | list[int]) -> np.ndarray:
    """
    read 2D array from raw file
    """
    if not len(lxs) == 3:
        raise ValueError(f"lxs must have 3 elements, you have lxs={lxs}")

    return np.fromfile(f, np.float64, np.prod(lxs[1:])).reshape(*lxs[1:], order="F")


def glow_aurmap(file: Path) -> xarray.Dataset:
    """
    read the auroral output from GLOW
    """

    lxs = simsize(file.parent)
    xg = grid(file.parent)
    dat = xarray.Dataset(coords={"wavelength": WAVELEN, "x2": xg["x2"][2:-2], "x3": xg["x3"][2:-2]})

    if not len(lxs) == 3:
        raise ValueError(f"lxs must have 3 elements, you have lxs={lxs}")

    with file.open("r") as f:
        raw = np.fromfile(f, np.float64, np.prod(lxs[1:]) * len(WAVELEN)).reshape(
            np.prod(lxs[1:]) * len(WAVELEN), order="F"
        )

    dat["rayleighs"] = (("wavelength", "x2", "x3"), raw)


def time(f) -> datetime:
    t = np.fromfile(f, np.float64, 4)
    return datetime(int(t[0]), int(t[1]), int(t[2])) + timedelta(hours=t[3])



#!/usr/bin/env python
"""
get Linux system info
"""

from __future__ import annotations
from configparser import ConfigParser
from pathlib import Path
import sys


def os_release() -> list[str]:
    """
    reads /etc/os-release with fallback to legacy methods

    Returns
    -------

    linux_names: list of str
        name(s) of operating system detect
    """

    if sys.platform != "linux":
        return []

    fn = Path("/etc/os-release")
    if not fn.is_file():
        if Path("/etc/redhat-release").is_file() or Path("/etc/centos-release").is_file():
            return ["rhel"]
        elif Path("/etc/debian_version").is_file():
            return ["debian"]
        elif Path("/etc/arch-version").is_file():
            return ["arch"]

    return parse_os_release("[all]" + fn.read_text())


def parse_os_release(txt: str) -> list[str]:
    """ parse /etc/os-release text """

    C = ConfigParser(inline_comment_prefixes=("#", ";"))
    C.read_string(txt)
    like = C["all"].get("ID_LIKE", fallback="")
    if not like:
        like = C["all"].get("ID", fallback="")
    like = like.strip('"').strip("'").split()

    return like


def get_package_manager(like: list[str] = None) -> str:
    if not like:
        like = os_release()
    if isinstance(like, str):
        like = [like]

    if {"centos", "rhel", "fedora"}.intersection(like):
        return "yum"
    elif {"debian", "ubuntu"}.intersection(like):
        return "apt"
    elif like == "arch":
        return "pacman"
    else:
        raise ValueError(
            f"Unknown ID_LIKE={like}, please file bug report or manually specify package manager"
        )



from __future__ import annotations
from pathlib import Path
import math

from .utils import get_cpu_count
from . import read


def count(path: Path, max_cpu: int) -> int:
    """get appropriate MPI image count for problem shape

    Parameters
    ----------
    path: pathlib.Path
        simsize file

    Returns
    -------
    count: int
        detect number of physical CPU
    """

    return max_mpi(read.simsize(path), max_cpu)


def max_mpi(size: tuple[int, ...], max_cpu: int) -> int:
    """
    goal is to find the highest x2 + x3 to maximum CPU core count
    """

    if len(size) != 3:
        raise ValueError("expected x1,x2,x3")

    if not max_cpu:
        max_cpu = get_cpu_count()

    if size[2] == 1:
        # 2D sim
        N = max_gcd(size[1], max_cpu)
    elif size[1] == 1:
        # 2D sim
        N = max_gcd(size[2], max_cpu)
    else:
        # 3D sim
        N = max_gcd2(size[1:], max_cpu)

    return N


def max_gcd(s: int, M: int) -> int:
    """
    find the Greatest Common Factor to evenly partition the simulation grid

    Output range is [M, 1]
    """

    if M < 1:
        raise ValueError("CPU count must be at least one")

    N = 1
    for i in range(M, 1, -1):
        N = max(math.gcd(s, i), N)
        if i < N:
            break

    return N


def max_gcd2(s: tuple[int, ...] | list[int], M: int) -> int:
    """
    find the Greatest Common Factor to evenly partition the simulation grid

    Output range is [M, 1]

    1. find factors of each dimension
    2. choose partition that yields highest CPU count usage
    """

    if len(s) != 2:
        raise ValueError("expected x2,x3")

    if M < 1:
        raise ValueError("CPU count must be at least one")

    f2 = [max_gcd(s[0], m) for m in range(M, 0, -1)]
    f3 = [max_gcd(s[1], m) for m in range(M, 0, -1)]

    N = 1
    for i in f2:
        for j in f3:
            if M >= i * j > N:
                # print(i,j)
                N = i * j
    return N



from __future__ import annotations
import typing as T
import numpy as np
import xarray

from . import write
from .config import datetime_range


def particles_BCs(cfg: dict[str, T.Any], xg: dict[str, T.Any]):
    """ write particle precipitation to disk """

    pg = precip_grid(cfg, xg)

    # %% CREATE PRECIPITATION INPUT DATA
    # Q: energy flux [mW m^-2]
    # E0: characteristic energy [eV]

    # did user specify on/off time? if not, assume always on.
    t0 = pg.time[0].values

    if "precip_startsec" in cfg:
        t = t0 + np.timedelta64(cfg["precip_startsec"])
        i_on = abs(pg.time - t).argmin().item()
    else:
        i_on = 0

    if "precip_endsec" in cfg:
        t = t0 + np.timedelta64(cfg["precip_endsec"])
        i_off = abs(pg.time - t).argmin().item()
    else:
        i_off = pg.time.size

    # NOTE: in future, E0 could be made time-dependent in config.nml as 1D array
    for i in range(i_on, i_off):
        pg["Q"][i, :, :] = precip_gaussian2d(pg, cfg["Qprecip"], cfg["Qprecip_background"])
        pg["E0"][i, :, :] = cfg["E0precip"]

    # %% CONVERT THE ENERGY TO EV
    # E0 = max(E0,0.100);
    # E0 = E0*1e3;

    # %% SAVE to files
    # LEAVE THE SPATIAL AND TEMPORAL INTERPOLATION TO THE
    # FORTRAN CODE IN CASE DIFFERENT GRIDS NEED TO BE TRIED.
    # THE EFIELD DATA DO NOT NEED TO BE SMOOTHED.

    write.precip(pg, cfg["precdir"], cfg["file_format"])


def precip_grid(cfg: dict[str, T.Any], xg: dict[str, T.Any]) -> xarray.Dataset:
    """CREATE PRECIPITATION CHARACTERISTICS data
    grid cells will be interpolated to grid, so 100x100 is arbitrary
    """

    lx2 = None
    lx3 = None

    llon = 100
    llat = 100
    # NOTE: cartesian-specific code
    for k in ("lx", "lxs"):
        if k in xg:
            _, lx2, lx3 = xg[k]
            break
    if lx2 == 1:
        llon = 1
    elif lx3 == 1:
        llat = 1

    if lx2 is None:
        raise ValueError("size data not in Efield grid")

    thetamin = xg["theta"].min()
    thetamax = xg["theta"].max()
    mlatmin = 90 - np.degrees(thetamax)
    mlatmax = 90 - np.degrees(thetamin)
    mlonmin = np.degrees(xg["phi"].min())
    mlonmax = np.degrees(xg["phi"].max())

    # add a 1% buff
    latbuf = 0.01 * (mlatmax - mlatmin)
    lonbuf = 0.01 * (mlonmax - mlonmin)

    pg = xarray.Dataset(
        {
            "Q": (("time", "mlon", "mlat"), np.zeros((len(cfg["time"]), llon, llat))),
            "E0": (("time", "mlon", "mlat"), np.zeros((len(cfg["time"]), llon, llat))),
        },
        coords={
            "time": datetime_range(cfg["time"][0], cfg["time"][0] + cfg["tdur"], cfg["dtprec"]),
            "mlat": np.linspace(mlatmin - latbuf, mlatmax + latbuf, llat),
            "mlon": np.linspace(mlonmin - lonbuf, mlonmax + lonbuf, llon),
        },
    )

    # %% disturbance extents
    # avoid divide by zero
    if "precip_latwidth" in cfg:
        pg.attrs["mlat_sigma"] = max(cfg["precip_latwidth"] * (mlatmax - mlatmin), 0.01)
    if "precip_lonwidth" in cfg:
        pg.attrs["mlon_sigma"] = max(cfg["precip_lonwidth"] * (mlonmax - mlonmin), 0.01)

    return pg


def precip_gaussian2d(pg: xarray.Dataset, Qpeak: float, Qbackground: float) -> np.ndarray:

    mlon_mean = pg.mlon.mean().item()
    mlat_mean = pg.mlat.mean().item()

    if "mlon_sigma" in pg.attrs and "mlat_sigma" in pg.attrs:
        Q = (
            Qpeak
            * np.exp(-((pg.mlon.values[:, None] - mlon_mean) ** 2) / (2 * pg.mlon_sigma ** 2))
            * np.exp(-((pg.mlat.values[None, :] - mlat_mean) ** 2) / (2 * pg.mlat_sigma ** 2))
        )
    elif "mlon_sigma" in pg.attrs:
        Q = Qpeak * np.exp(-((pg.mlon.values[:, None] - mlon_mean) ** 2) / (2 * pg.mlon_sigma ** 2))
    elif "mlat_sigma" in pg.attrs:
        Q = Qpeak * np.exp(-((pg.mlat.values[None, :] - mlat_mean) ** 2) / (2 * pg.mlat_sigma ** 2))
    else:
        raise LookupError("precipation must be defined in latitude, longitude or both")

    Q[Q < Qbackground] = Qbackground

    return Q



"""
using MSIS Fortran exectuable from Python
"""

from __future__ import annotations
import xarray
from pathlib import Path
import numpy as np
import subprocess
import logging
import typing as T
import h5py

from . import cmake


def msis_setup(p: dict[str, T.Any], xg: dict[str, T.Any]) -> xarray.Dataset:
    """
    calls MSIS Fortran executable msis_setup--builds if not present

    [f107a, f107, ap] = activ
    """

    msis_exe = cmake.build_gemini3d(Path("msis_setup"))

    alt_km = xg["alt"] / 1e3
    # % CONVERT DATES/TIMES/INDICES INTO MSIS-FRIENDLY FORMAT
    t0 = p["time"][0]
    doy = int(t0.strftime("%j"))
    UTsec0 = t0.hour * 3600 + t0.minute * 60 + t0.second + t0.microsecond / 1e6
    # censor BELOW-ZERO ALTITUDES SO THAT THEY DON'T GIVE INF
    alt_km[alt_km <= 0] = 1
    # %% CREATE INPUT FILE FOR FORTRAN PROGRAM
    msis_infile = p.get("msis_infile", p["indat_size"].parent / "msis_setup_in.h5")
    msis_outfile = p.get("msis_outfile", p["indat_size"].parent / "msis_setup_out.h5")

    with h5py.File(msis_infile, "w") as f:
        f["/doy"] = doy
        f["/UTsec"] = UTsec0
        f["/f107a"] = p["f107a"]
        f["/f107"] = p["f107"]
        f["/Ap"] = [p["Ap"]] * 7
        # astype(float32) just to save disk I/O time/space
        f["/glat"] = xg["glat"].astype(np.float32)
        f["/glon"] = xg["glon"].astype(np.float32)
        f["/alt"] = alt_km.astype(np.float32)
    # %% run MSIS
    args = [str(msis_infile), str(msis_outfile)]

    if "msis_version" in p:
        args.append(str(p["msis_version"]))
    cmd = [str(msis_exe)] + args
    logging.info(" ".join(cmd))
    ret = subprocess.run(cmd, text=True, cwd=msis_exe.parent)

    if ret.returncode == 20:
        raise RuntimeError("Need to compile with 'cmake -Dmsis20=true'")
    if ret.returncode != 0:
        raise RuntimeError(f"MSIS failed to run: {ret.stdout}")
    # %% load MSIS output
    alt1 = alt_km[:, 0, 0]
    glat1 = xg["glat"][0, :, 0]
    glon1 = xg["glon"][0, 0, :]
    atmos = xarray.Dataset(coords={"alt_km": alt1, "glat": glat1, "glon": glon1})

    with h5py.File(msis_outfile, "r") as f:
        for k in {"nO", "nN2", "nO2", "Tn", "nN", "nH"}:
            atmos[k] = (("alt_km", "glat", "glon"), f[f"/{k}"][:])

    # Mitra, 1968
    atmos["nNO"] = 0.4 * np.exp(-3700 / atmos["Tn"]) * atmos["nO2"] + 5e-7 * atmos["nO"]

    return atmos



"""
test msis
"""

import pytest
import numpy as np
from pytest import approx
from datetime import datetime

import gemini3d.msis as gp


@pytest.mark.parametrize("version", [0, 20])
def test_msis(version, tmp_path):
    cfg = {
        "time": [datetime(2015, 1, 2, 12)],
        "f107": 100.0,
        "f107a": 100.0,
        "Ap": 4,
        "msis_version": version,
        "indat_size": tmp_path / "inputs/simsize.h5",
    }

    (tmp_path / "inputs").mkdir(exist_ok=True)

    lx = (4, 2, 3)

    glon, alt, glat = np.meshgrid(
        np.linspace(-147, -145, lx[1]),
        np.linspace(100e3, 200e3, lx[0]),
        np.linspace(65, 66, lx[2]),
    )
    xg = {
        "glat": glat,
        "glon": glon,
        "lx": lx,
        "alt": alt,
    }

    try:
        atmos = gp.msis_setup(cfg, xg)
    except RuntimeError as e:
        if "-Dmsis20=" in str(e):
            pytest.skip("MSIS 2.0 wasn't available")
        raise

    assert atmos["Tn"].shape == lx

    truth = {
        0: {
            "nO": 5.8859558e17,
            "nN2": 9.2224699e18,
            "nO2": 2.2804600e18,
            "Tn": 185.59,
            "nN": 2.0134611e11,
            "nNO": 2.9630136e11,
            "nH": 2.3853709e13,
        },
        20: {
            "nO": 4.3219809e17,
            "nN2": 7.6246557e18,
            "nO2": 1.9486360e18,
            "Tn": 189.72,
            "nN": 4.4032811e10,
            "nNO": 2.1874142e11,
            "nH": 1.7495320e13,
        },
    }

    # pick an arbitrary 3D location
    a1 = atmos.isel({"alt_km": 0, "glat": 1, "glon": 2})

    for k in a1.data_vars:
        assert a1[k] == approx(
            truth[version][k],
            rel=1e-5,
        )



import shutil
from datetime import datetime
import pytest
from pathlib import Path
import sys

try:
    import gemini3d.plot
except ImportError as e:
    pytest.skip(f"Matplotlib missing {e}", allow_module_level=True)

R = Path(__file__).parent / "data"


@pytest.mark.skipif(sys.version_info < (3, 8), reason="Python >= 3.8 needed for copytree test")
@pytest.mark.parametrize(
    "name",
    [
        "2dew_glow",
        "2dns_glow",
        "3d_glow",
    ],
)
def test_plot(name, tmp_path):

    # get files if needed
    try:
        test_dir = gemini3d.web.download_and_extract(name, R)
    except ConnectionError as e:
        pytest.skip(f"failed to download reference data {e}")

    shutil.copytree(test_dir, tmp_path, dirs_exist_ok=True)
    gemini3d.plot.frame(tmp_path, datetime(2013, 2, 20, 5), saveplot_fmt="png")

    plot_files = sorted((tmp_path / "plots").glob("*.png"))

    assert len(plot_files) == 10



"""
Compile HDF5 library

Be sure environment variables are set for your desired compiler.
Use the full compiler path if it's not getting the right compiler.

* FC: Fortran compiler name or path
* CC: C compiler name or path
"""

from __future__ import annotations
import typing as T
import sys
import os
import subprocess
import shutil
import argparse
import tempfile
import json
import importlib.resources
from pathlib import Path

from . import cmake
from .utils import get_cpu_count
from .web import url_retrieve, extract_tar, git_download

# ========= user parameters ======================
BUILDDIR = "build"

HDF5_DIR = "hdf5"
LAPACK_DIR = "lapack"
SCALAPACK_DIR = "scalapack"
MUMPS_DIR = "mumps"

# ========= end of user parameters ================


def cli():
    p = argparse.ArgumentParser(
        description="Compiles prerequisite libraries for Gemini (or other programs)"
    )
    p.add_argument(
        "compiler", help="compiler to build libraries for", choices=["gcc", "intel", "ibmxl"]
    )
    p.add_argument(
        "libs",
        help="libraries to compile",
        choices=["hdf5", "lapack", "mumps", "netcdf", "openmpi", "openmpi3", "scalapack"],
        nargs="+",
    )
    p.add_argument("-prefix", help="top-level directory to install libraries under")
    p.add_argument(
        "-workdir",
        help="top-level directory to build under (can be deleted when done)",
        default=tempfile.gettempdir(),
    )
    p.add_argument("-reuse", help="reuse existing build (not usually done)", action="store_true")
    p.add_argument(
        "-n",
        "--dryrun",
        help="download code and configure but don't actually build (for testing only)",
        action="store_true",
    )
    P = p.parse_args()

    main(P.compiler, P.libs, P.prefix, P.workdir, P.reuse, P.dryrun)


def main(compiler: str, libs: list[str], prefix: str, workdir: str, reuse: bool, dryrun: bool):

    prefix = prefix if prefix else f"~/lib_{compiler}"

    dirs = {
        "prefix": Path(prefix).expanduser().resolve(),
        "workdir": Path(workdir).expanduser().resolve(),
    }

    setup_libs(libs, dirs, compiler, wipe=not reuse, dryrun=dryrun)


def setup_libs(
    libs: list[str],
    dirs: dict[str, Path],
    compiler: str,
    *,
    wipe: bool,
    dryrun: bool = False,
):

    if compiler == "gcc":
        env = gcc_compilers()
    elif compiler == "intel":
        env = intel_compilers()
    elif compiler == "ibmxl":
        env = ibmxl_compilers()
    else:
        raise ValueError(f"unknown compiler {compiler}")

    if "hdf5" in libs:
        hdf5(dirs, env=env)
    if "netcdf" in libs:
        netcdf_c(dirs, env=env, wipe=wipe, dryrun=dryrun)
        netcdf_fortran(dirs, env=env, wipe=wipe, dryrun=dryrun)

    # Note: OpenMPI needs to be before scalapack and mumps
    if "openmpi" in libs:
        openmpi(dirs, env=env, version="", dryrun=dryrun)
    elif "openmpi3" in libs:
        openmpi(dirs, env=env, version="3", dryrun=dryrun)

    if "lapack" in libs:
        lapack(wipe, dirs, env=env, dryrun=dryrun)
    if "scalapack" in libs:
        scalapack(wipe, dirs, env=env, dryrun=dryrun)
    if "mumps" in libs:
        mumps(wipe, dirs, env=env, dryrun=dryrun)

    if not dryrun:
        print("Installed", libs, "under", dirs["prefix"])


def netcdf_c(
    dirs: dict[str, Path], env: T.Mapping[str, str], wipe: bool = False, dryrun: bool = False
):
    """build and install NetCDF-C"""

    install_dir = dirs["prefix"] / "netcdf"
    source_dir = dirs["workdir"] / "netcdf-c"
    build_dir = source_dir / BUILDDIR

    git_json(source_dir, "netcdf-c")

    hdf5_root = dirs["prefix"] / HDF5_DIR
    if hdf5_root.is_dir():
        lib_args = [f"-DHDF5_ROOT={hdf5_root.as_posix()}"]
    else:
        lib_args = []

    if not cmake.find_library("HDF5 COMPONENTS C Fortran", lib_args, env):
        raise RuntimeError("Please install HDF5 before NetCDF4")

    c_args = [
        f"-DCMAKE_INSTALL_PREFIX:PATH={install_dir}",
        "-DCMAKE_BUILD_TYPE:STRING=Release",
        "-DBUILD_SHARED_LIBS:BOOL=ON",
        "-DENABLE_PARALLEL4:BOOL=OFF",
        "-DENABLE_PNETCDF:BOOL=OFF",
        "-DBUILD_UTILITIES:BOOL=OFF",
        "-DENABLE_TESTS:BOOL=off",
        "-DBUILD_TESTING:BOOL=OFF",
        "-DENABLE_HDF4:BOOL=OFF",
        "-DUSE_DAP:BOOL=off",
        "-DENABLE_DAP:BOOL=OFF",
        "-DENABLE_DAP2:BOOL=OFF",
        "-DENABLE_DAP4:BOOL=OFF",
    ]
    cmake.build(
        source_dir,
        build_dir,
        wipe=wipe,
        env=env,
        run_test=False,
        dryrun=dryrun,
        config_args=c_args + lib_args,
    )


def netcdf_fortran(
    dirs: dict[str, Path], env: T.Mapping[str, str], wipe: bool = False, dryrun: bool = False
):
    """build and install NetCDF-Fortran"""

    install_dir = dirs["prefix"] / "netcdf"
    source_dir = dirs["workdir"] / "netcdf-fortran"
    build_dir = source_dir / BUILDDIR

    git_json(source_dir, "netcdf-fortran")

    # NetCDF-Fortran does not yet use NetCDF_ROOT
    if sys.platform == "linux":
        netcdf_c = install_dir / "lib/libnetcdf.so"
    elif sys.platform == "win32":
        print(
            "NetCDF4 on MSYS2 may not work, see https://github.com/Unidata/netcdf-c/issues/554",
            file=sys.stderr,
        )
        netcdf_c = install_dir / "bin/libnetcdf.dll"
    elif sys.platform == "darwin":
        netcdf_c = install_dir / "lib/libnetcdf.dylib"
    else:
        raise EnvironmentError(
            f"please open a GitHub Issue for your operating system {sys.platform}"
        )

    hdf5_root = dirs["prefix"] / HDF5_DIR
    if hdf5_root.is_dir():
        lib_args = [f"-DHDF5_ROOT={hdf5_root.as_posix()}"]
    else:
        lib_args = []

    if not cmake.find_library("HDF5 COMPONENTS C Fortran", lib_args, env):
        raise RuntimeError("Please install HDF5 before NetCDF4")

    patch = [
        f"-DNETCDF_C_LIBRARY:FILEPATH={netcdf_c}",
        f"-DNETCDF_INCLUDE_DIR:PATH={install_dir / 'include'}",
    ]
    f_args = patch + [
        f"-DNetCDF_ROOT:PATH={install_dir}",
        f"-DCMAKE_INSTALL_PREFIX:PATH={install_dir}",
        "-DCMAKE_BUILD_TYPE:STRING=Release",
        "-DBUILD_SHARED_LIBS:BOOL=ON",
        "-DBUILD_UTILITIES:BOOL=OFF",
        "-DENABLE_TESTS:BOOL=off",
        "-DBUILD_EXAMPLES:BOOL=OFF",
    ]
    cmake.build(
        source_dir, build_dir, wipe=wipe, env=env, run_test=False, dryrun=dryrun, config_args=f_args
    )


def hdf5(dirs: dict[str, Path], env: T.Mapping[str, str]):
    """build and install HDF5
    some systems have broken libz and so have trouble extracting tar.bz2 from Python.
    To avoid this, we git clone the release instead.
    """

    use_cmake = True
    name = "hdf5"
    install_dir = dirs["prefix"] / name
    source_dir = dirs["workdir"] / name
    build_dir = source_dir / BUILDDIR

    git_json(source_dir, "hdf5")

    if use_cmake or os.name == "nt":
        # works for Intel oneAPI on Windows and many other systems/compilers.
        # works for Make or Ninja in general.
        cmd0 = [
            "cmake",
            f"-S{source_dir}",
            f"-DCMAKE_INSTALL_PREFIX={install_dir}",
            "-DHDF5_GENERATE_HEADERS:BOOL=false",
            "-DHDF5_DISABLE_COMPILER_WARNINGS:BOOL=true",
            "-DBUILD_SHARED_LIBS:BOOL=false",
            "-DCMAKE_BUILD_TYPE=Release",
            "-DHDF5_BUILD_FORTRAN:BOOL=true",
            "-DHDF5_BUILD_CPP_LIB:BOOL=false",
            "-DHDF5_BUILD_TOOLS:BOOL=false",
            "-DBUILD_TESTING:BOOL=false",
            "-DHDF5_BUILD_EXAMPLES:BOOL=false",
        ]

        cmd1 = ["cmake", "--build", str(build_dir), "--parallel"]

        cmd2 = ["cmake", "--install", str(build_dir)]

        # this old "cmake .." style command is necessary due to bugs with
        # HDF5 (including 1.10.7) CMakeLists:
        #   CMake Error at config/cmake/HDF5UseFortran.cmake:205 (file):
        #   file failed to open for reading (No such file or directory):
        #   C:/Users/micha/AppData/Local/Temp/hdf5/build/pac_fconftest.out.
        build_dir.mkdir(exist_ok=True)
        subprocess.check_call(cmd0, cwd=build_dir, env=env)
    else:
        cmd0 = [
            "./configure",
            f"--prefix={install_dir}",
            "--enable-fortran",
            "--enable-build-mode=production",
        ]
        cmd1 = ["make", "-j"]
        cmd2 = ["make", "-j", "install"]
        subprocess.check_call(cmd0, cwd=source_dir, env=env)

    subprocess.check_call(cmd1, cwd=source_dir)
    subprocess.check_call(cmd2, cwd=source_dir)


def openmpi(dirs: dict[str, Path], env: T.Mapping[str, str], version: str, dryrun: bool = False):
    """ build and install OpenMPI """
    if os.name == "nt":
        raise EnvironmentError(
            """
OpenMPI is not available in native Windows.
Use MPI on Windows via any of (choose one):

* Windows Subsystem for Linux
* MS-MPI with MSYS2: https://www.scivision.dev/windows-mpi-msys2/
* Intel oneAPI with IntelMPI: https://www.scivision.dev/intel-oneapi-fortran-install/
* Cygwin
"""
        )

    jmeta = json.loads(importlib.resources.read_text(__package__, "libraries.json"))
    version = jmeta[f"openmpi{version}"]["tag"]

    mpi_dir = f"openmpi-{version}"
    install_dir = dirs["prefix"] / mpi_dir
    source_dir = dirs["workdir"] / mpi_dir

    tar_name = f"openmpi-{version}.tar.bz2"
    tarfn = dirs["workdir"] / tar_name

    url = f"https://download.open-mpi.org/release/open-mpi/v{version[:3]}/{tar_name}"
    url_retrieve(url, tarfn)
    extract_tar(tarfn, source_dir)

    cmd = [
        "./configure",
        f"--prefix={install_dir}",
        f"CC={env['CC']}",
        f"CXX={env['CXX']}",
        f"FC={env['FC']}",
    ]

    subprocess.check_call(cmd, cwd=source_dir, env=env)

    Njobs = get_cpu_count()

    cmd = ["make", "-C", str(source_dir), "-j", str(Njobs), "install"]

    if dryrun:
        print("DRYRUN: would have run\n", " ".join(cmd))
        return None

    subprocess.check_call(cmd)


def lapack(wipe: bool, dirs: dict[str, Path], env: T.Mapping[str, str], dryrun: bool = False):
    """ build and insall Lapack """
    install_dir = dirs["prefix"] / LAPACK_DIR
    source_dir = dirs["workdir"] / LAPACK_DIR
    build_dir = source_dir / BUILDDIR

    git_json(source_dir, "lapack")

    args = ["-Dautobuild:BOOL=off", f"-DCMAKE_INSTALL_PREFIX:PATH={install_dir}"]
    cmake.build(source_dir, build_dir, wipe=wipe, env=env, dryrun=dryrun, config_args=args)


def scalapack(wipe: bool, dirs: dict[str, Path], env: T.Mapping[str, str], dryrun: bool = False):
    """ build and install Scalapack """
    install_dir = dirs["prefix"] / SCALAPACK_DIR
    source_dir = dirs["workdir"] / SCALAPACK_DIR
    build_dir = source_dir / BUILDDIR

    git_json(source_dir, "scalapack")

    lapack_root = dirs["prefix"] / LAPACK_DIR
    lib_args = [f"-DLAPACK_ROOT={lapack_root.as_posix()}"]

    if not cmake.find_library("LAPACK", lib_args, env):
        lapack(wipe, dirs, env)

    args = [
        "-Dautobuild:BOOL=off",
        f"-DCMAKE_INSTALL_PREFIX:PATH={install_dir}",
    ]
    cmake.build(
        source_dir, build_dir, wipe=wipe, env=env, dryrun=dryrun, config_args=args + lib_args
    )


def mumps(wipe: bool, dirs: dict[str, Path], env: T.Mapping[str, str], dryrun: bool = False):
    """ build and install Mumps """
    install_dir = dirs["prefix"] / MUMPS_DIR
    source_dir = dirs["workdir"] / MUMPS_DIR
    build_dir = source_dir / BUILDDIR

    scalapack_lib = dirs["prefix"] / SCALAPACK_DIR
    lapack_lib = dirs["prefix"] / LAPACK_DIR

    git_json(source_dir, "mumps")

    if env["FC"] == "ifort":
        lib_args = []
    else:
        lib_args = [
            f"-DSCALAPACK_ROOT:PATH={scalapack_lib.as_posix()}",
            f"-DLAPACK_ROOT:PATH={lapack_lib.as_posix()}",
        ]

    if not cmake.find_library("LAPACK", lib_args, env):
        lapack(wipe, dirs, env)
    if not cmake.find_library("SCALAPACK", lib_args, env):
        scalapack(wipe, dirs, env)

    args = ["-Dautobuild:BOOL=off", f"-DCMAKE_INSTALL_PREFIX:PATH={install_dir}"]
    cmake.build(
        source_dir, build_dir, wipe=wipe, env=env, dryrun=dryrun, config_args=args + lib_args
    )


def git_json(path: Path, name: str):
    jmeta = json.loads(importlib.resources.read_text(__package__, "libraries.json"))

    git_download(path, jmeta[name]["git"], tag=jmeta[name].get("tag"))


def get_compilers(compiler_name: str, **kwargs) -> T.Mapping[str, str]:
    """get paths to compilers

    Parameters
    ----------

    compiler_name: str
        arbitrary string naming compiler--to give useful error message when compiler not found.
    """
    env = os.environ

    for k, v in kwargs.items():
        c = env.get(k, "")
        if v not in c:
            c = shutil.which(v)
        if not c:
            raise FileNotFoundError(
                f"Compiler {compiler_name} was not found: {k}."
                " Did you load the compiler shell environment first?"
            )
        env.update({k: c})

    return env


def gcc_compilers() -> T.Mapping[str, str]:
    return get_compilers("GNU", FC="gfortran", CC="gcc", CXX="g++")


def intel_compilers() -> T.Mapping[str, str]:
    return get_compilers(
        "Intel",
        FC="ifort",
        CC="icl" if os.name == "nt" else "icc",
        CXX="icl" if os.name == "nt" else "icpc",
    )


def ibmxl_compilers() -> T.Mapping[str, str]:
    return get_compilers("IBM XL", FC="xlf", CC="xlc", CXX="xlc++")


if __name__ == "__main__":
    cli()



from __future__ import annotations
import numpy as np
import math

pi = math.pi


def geomag2geog(thetat: np.ndarray, phit: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """ geomagnetic to geographic """

    # FIXME: this is for year 1985, see Schmidt spherical harmonic in MatGemini
    thetan = math.radians(11)
    phin = math.radians(289)

    # enforce phit = [0,2pi]
    phit = np.remainder(phit, 2 * pi)

    # thetag2p=acos(cos(thetat).*cos(thetan)-sin(thetat).*sin(thetan).*cos(phit));
    thetag2p = np.arccos(
        np.cos(thetat) * np.cos(thetan) - np.sin(thetat) * np.sin(thetan) * np.cos(phit)
    )

    beta = np.arccos(
        (np.cos(thetat) - np.cos(thetag2p) * np.cos(thetan)) / (np.sin(thetag2p) * np.sin(thetan))
    )

    phig2 = np.zeros_like(phit)

    i = phit > pi
    phig2[i] = phin - beta[i]

    i = phit <= pi
    phig2[i] = phin + beta[i]

    i = phig2 < 0
    phig2[i] = phig2[i] + 2 * pi

    i = phig2 >= 2 * pi
    phig2[i] = phig2[i] - 2 * pi

    thetag2 = pi / 2 - thetag2p
    lat = np.degrees(thetag2)
    lon = np.degrees(phig2)

    return lat, lon


def geog2geomag(lat: np.ndarray, lon: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """ geographic to geomagnetic """

    # FIXME: this is for year 1985, see Schmidt spherical harmonic in MatGemini
    thetan = math.radians(11)
    phin = math.radians(289)

    lat = np.atleast_1d(lat)
    lon = np.atleast_1d(lon)

    # enforce [0,360] longitude
    lon = lon % 360

    thetagp = pi / 2 - np.radians(lat)
    phig = np.radians(lon)

    thetat = np.arccos(
        np.cos(thetagp) * np.cos(thetan) + np.sin(thetagp) * np.sin(thetan) * np.cos(phig - phin)
    )
    argtmp = (np.cos(thetagp) - np.cos(thetat) * np.cos(thetan)) / (np.sin(thetat) * np.sin(thetan))
    alpha = np.arccos(max(min(argtmp, 1), -1))
    phit = np.empty(lat.shape)

    i = ((phin > phig) & ((phin - phig) > pi)) | ((phin < phig) & ((phig - phin) < pi))
    phit[i] = pi - alpha[i]
    phit[~i] = alpha[~i] + pi

    return thetat, phit


def geog2UEN(alt, glon, glat, thetactr, phictr):
    """
    Converts a set of glon,glat into magnetic up, north, east coordinates.
    thetactr and phictr are the magnetic coordinates of the center of hte region of interest.
    They can be computed from geog2geomag.
    """

    # %% UPWARD DISTANCE
    Re = 6370e3
    z = alt

    # Convert to geomganetic coordinates
    theta, phi = geog2geomag(glat, glon)

    # Convert to northward distance in meters
    gamma2 = theta - thetactr  # southward magnetic angular distance
    gamma2 = -gamma2  # convert to northward angular distance
    y = gamma2 * Re

    gamma1 = phi - phictr  # eastward angular distance
    x = Re * np.sin(thetactr) * gamma1

    return z, x, y


def UEN2geog(z, x, y, thetactr, phictr):
    """
    converts magnetic up, north, east coordinates into geographic coordinates.
    """

    # UPWARD DISTANCE
    Re = 6370e3
    alt = z

    # Northward angular distance
    gamma2 = y / Re  # must retain the sign of x3
    theta = thetactr - gamma2  # minus because distance north is against theta's direction

    # Eastward angular distance
    gamma1 = x / Re / np.sin(thetactr)
    # must retain the sign of x2, just use theta of center of grid
    phi = phictr + gamma1

    # Now convert the magnetic to geographic using our simple transformation
    [glat, glon] = geomag2geog(theta, phi)

    return alt, glon, glat



"""
these test that PyGemini generates inputs that match expectations
"""

from datetime import datetime
import logging
import pytest
from pathlib import Path

import gemini3d.web
import gemini3d.write as write
import gemini3d.grid
from gemini3d.compare import compare_all, compare_grid, compare_Efield, compare_precip
import gemini3d.read
import gemini3d.model
import gemini3d.find
from gemini3d.efield import Efield_BCs
from gemini3d.particles import particles_BCs

R = Path(__file__).parent / "data"


@pytest.mark.parametrize("name", ["2dew_fang"])
def test_file_time(name):
    # get files if needed
    try:
        test_dir = gemini3d.web.download_and_extract(name, R)
    except ConnectionError as e:
        pytest.skip(f"failed to download reference data {e}")

    t0 = datetime(2013, 2, 20, 5, 0, 0)

    file = gemini3d.find.frame(test_dir, time=t0)
    time = gemini3d.read.time(file)
    assert time == t0


@pytest.mark.parametrize(
    "name,file_format",
    [("2dew_fang", "h5"), ("3d_fang", "h5"), ("2dew_fang", "nc"), ("3d_fang", "nc")],
)
def test_grid(name, file_format, tmp_path):

    if file_format == "nc":
        pytest.importorskip("netCDF4")

    # get files if needed
    try:
        test_dir = gemini3d.web.download_and_extract(name, R)
    except ConnectionError as e:
        pytest.skip(f"failed to download reference data {e}")

    # setup new test data
    cfg = gemini3d.read.config(test_dir)
    xg = gemini3d.grid.cart3d(cfg)

    # path patch
    cfg["out_dir"] = tmp_path
    cfg["indat_size"] = cfg["out_dir"] / cfg["indat_size"]
    cfg["indat_grid"] = cfg["out_dir"] / cfg["indat_grid"]

    write.grid(cfg, xg, file_format=file_format)

    errs = compare_grid(cfg["out_dir"], test_dir, file_format=file_format)

    assert errs == 0, f"grid mismatch {cfg['out_dir']}  {test_dir}"


@pytest.mark.parametrize(
    "name,file_format", [("2dew_fang", "h5"), ("2dns_fang", "h5"), ("3d_fang", "h5")]
)
def test_Efield(name, file_format, tmp_path):

    # get files if needed
    try:
        test_dir = gemini3d.web.download_and_extract(name, R)
    except ConnectionError as e:
        pytest.skip(f"failed to download reference data {e}")

    cfg = gemini3d.read.config(test_dir)
    xg = gemini3d.read.grid(test_dir)

    # patch paths
    cfg["out_dir"] = tmp_path
    E0dir = cfg["E0dir"]
    cfg["E0dir"] = cfg["out_dir"] / cfg["E0dir"]
    Efield_BCs(cfg, xg)
    errs = compare_Efield(
        cfg["time"], cfg["E0dir"], refdir=test_dir / E0dir, plot=False, file_format=file_format
    )

    assert errs == 0, f"Efield mismatch {cfg['out_dir']}  {test_dir}"


@pytest.mark.parametrize(
    "name,file_format", [("2dew_fang", "h5"), ("2dns_fang", "h5"), ("3d_fang", "h5")]
)
def test_precip(name, file_format, tmp_path):

    # get files if needed
    try:
        test_dir = gemini3d.web.download_and_extract(name, R)
    except ConnectionError as e:
        pytest.skip(f"failed to download reference data {e}")

    cfg = gemini3d.read.config(test_dir)
    xg = gemini3d.read.grid(test_dir)

    # patch paths
    cfg["out_dir"] = tmp_path
    precdir = cfg["precdir"]
    cfg["precdir"] = cfg["out_dir"] / cfg["precdir"]

    particles_BCs(cfg, xg)

    errs = compare_precip(
        cfg["time"], cfg["precdir"], refdir=test_dir / precdir, plot=False, file_format=file_format
    )

    assert errs == 0, f"precipitation mismatch {cfg['out_dir']}  {test_dir}"


@pytest.mark.parametrize(
    "name,file_format",
    [
        ("2dew_eq", "h5"),
        ("2dew_fang", "h5"),
        ("2dew_glow", "h5"),
        ("2dns_eq", "h5"),
        ("2dns_fang", "h5"),
        ("2dns_glow", "h5"),
        ("3d_eq", "h5"),
        ("3d_fang", "h5"),
        ("3d_glow", "h5"),
    ],
)
def test_runner(name, file_format, tmp_path):

    out_dir = tmp_path
    # get files if needed
    try:
        test_dir = gemini3d.web.download_and_extract(name, R)
    except ConnectionError as e:
        pytest.skip(f"failed to download reference data {e}")

    # setup new test data
    params = gemini3d.read.config(test_dir)

    params["file_format"] = file_format
    params["out_dir"] = out_dir

    for k in ("indat_file", "indat_size", "indat_grid"):
        params[k] = params[k].with_suffix("." + file_format)

    # patch eq_dir to use reference data
    if "eq_dir" in params:
        eq_dir = test_dir.parent / params["eq_dir"].name
        if eq_dir.is_dir():
            print(f"Using {eq_dir} for equilibrium data")
        params["eq_dir"] = eq_dir

    # %% generate initial condition files
    gemini3d.model.setup(params, out_dir)

    # %% check generated files
    errs = compare_all(
        params["out_dir"], refdir=test_dir, only="in", plot=False, file_format=file_format
    )

    if errs:
        for err, v in errs.items():
            logging.error(f"compare:{err}: {v} errors")
        raise ValueError(f"compare_input: new generated inputs do not match reference for: {name}")



"""
installs Gemini3D prerequisite libraries for:

* Linux: CentOS, Debian, Ubuntu, Windows Subsystem for Linux
* MacOS: Homebrew
* Windows: MSYS2, Cygwin

assumes GCC/Gfortran
"""

from __future__ import annotations
import subprocess
import sys
from argparse import ArgumentParser
import shutil

PKG = {
    "yum": [
        "epel-release",
        "gcc-gfortran",
        "MUMPS-openmpi-devel",
        "lapack-devel",
        "scalapack-openmpi-devel",
        "openmpi-devel",
        "hdf5-devel",
    ],
    "apt": [
        "gfortran",
        "libmumps-dev",
        "liblapack-dev",
        "libscalapack-mpi-dev",
        "libopenmpi-dev",
        "openmpi-bin",
        "libhdf5-dev",
    ],
    "pacman": ["gcc-fortran", "lapack", "openmpi", "hdf5"],
    "brew": ["gcc", "ninja", "cmake", "lapack", "scalapack", "openmpi", "hdf5"],
    "cygwin": ["gcc-fortran", "liblapack-devel", "libopenmpi-devel"],
    "msys": [
        "mingw-w64-x86_64-gcc-fortran",
        "mingw-w64-x86_64-hdf5",
        "mingw-w64-x86_64-lapack",
        "mingw-w64-x86_64-scalapack",
        "mingw-w64-x86_64-mumps",
    ],
}


def main(package_manager: str):

    cmd: list[str] = []

    if sys.platform == "linux":

        if not package_manager:
            from gemini3d.linux_info import get_package_manager

            package_manager = get_package_manager()

        if package_manager == "yum":
            subprocess.run(["sudo", "yum", "--assumeyes", "updateinfo"])
            cmd = ["sudo", "yum", "install"] + PKG["yum"]
        elif package_manager == "apt":
            subprocess.run(["sudo", "apt", "update", "--yes"])
            cmd = ["sudo", "apt", "install"] + PKG["apt"]
        elif package_manager == "pacman":
            subprocess.run(["sudo", "pacman", "-S", "--needed"] + PKG["pacman"])
        else:
            raise ValueError(
                f"Unknown package manager {package_manager}, try installing the prereqs manually"
            )
    elif sys.platform == "darwin":
        if not shutil.which("brew"):
            raise SystemExit(
                "We assume Homebrew is available, need to manually install a Fortran compiler otherwise."
            )
        cmd = ["brew", "install"] + PKG["brew"]
        # autobuild Mumps, it's much faster
    elif sys.platform == "cygwin":
        cmd = ["setup-x86_64.exe", "-P"] + PKG["cygwin"]
    elif sys.platform == "win32":
        if not shutil.which("pacman"):
            raise SystemExit("Windows Subsystem for Linux or MSYS2 is recommended.")
        # assume MSYS2
        cmd = ["pacman", "-S", "--needed"] + PKG["msys"]
    else:
        raise ValueError(f"unknown platform {sys.platform}")

    print(" ".join(cmd))
    ret = subprocess.run(cmd)

    raise SystemExit(ret.returncode)


if __name__ == "__main__":
    p = ArgumentParser()
    p.add_argument(
        "package_manager",
        help="specify package manager e.g. apt, yum",
        choices=list(PKG.keys()),
        nargs="?",
    )
    P = p.parse_args()

    main(P.package_manager)



from __future__ import annotations
from pathlib import Path
from matplotlib.figure import Figure
import typing as T
from datetime import datetime
import logging
import numpy as np
import xarray

from .. import read
from .. import find
from .vis import grid2plotfun
from ..config import datetime_range


PARAMS = ["ne", "v1", "Ti", "Te", "J1", "v2", "v3", "J2", "J3", "Phi"]


def grid(direc: Path, only: list[str] = None, saveplot_fmt: str = None):
    """plot 3D grid

    Parameters
    ----------

    direc: pathlib.Path
        top-level path of simulation grid
    """

    direc = Path(direc).expanduser()

    xg = read.grid(direc)

    if only is None:
        only = ["basic", "alt", "geog"]
        if xg["lxs"].prod() < 10000:
            # gets extremely slow if too many points
            only.append("ecef")

    # %% x1, x2, x3
    if "basic" in only:
        fg = basic(xg)
        stitle(fg, xg)
        save_fig(fg, direc, "grid-basic")

    # %% detailed altitude plot
    if "alt" in only:
        fg = altitude_grid(xg)
        save_fig(fg, direc, "grid-altitude")

    # %% ECEF surface
    if "ecef" in only:
        fg = Figure()
        ax = fg.gca(projection="3d")
        ax.scatter(xg["x"], xg["y"], xg["z"])

        ax.set_xlabel("x [m]")
        ax.set_ylabel("y [m]")
        ax.set_zlabel("z [m]")

        stitle(fg, xg, "ECEF")
        save_fig(fg, direc, "grid-ecef")

    # %% lat lon map
    if "geog" in only:
        fg = grid_geog(xg)
        save_fig(fg, direc, "grid-geog")


def grid_geog(xg: dict[str, T.Any]) -> Figure:
    """
    plots grid in geographic map
    """

    fig = Figure()

    glon = xg["glon"]
    glat = xg["glat"]

    try:
        import cartopy

        proj = cartopy.crs.PlateCarree()  # arbitrary

        ax = fig.gca(projection=proj)
        ax.add_feature(cartopy.feature.LAND)
        ax.add_feature(cartopy.feature.OCEAN)
        ax.add_feature(cartopy.feature.COASTLINE)
        ax.add_feature(cartopy.feature.BORDERS, linestyle=":")
        ax.gridlines(draw_labels=True, x_inline=False, y_inline=False)
        ax.scatter(glon, glat, transform=proj)
    except ImportError as e:
        logging.error(e)
        ax = fig.gca()
        ax.scatter(glon, glat)

    ax.set_xlabel("geographic longitude")
    ax.set_ylabel("geographic latitude")
    stitle(fig, xg, "glat, glon")

    return fig


def Efield(direc: Path):
    """plot input E-field

    Parameters
    ----------

    direc: pathlib.Path
        top-level simulation directory
    """

    direc = Path(direc).expanduser()

    cfg = read.config(direc)
    path = find.inputs(direc, cfg.get("E0dir"))
    if not path:
        raise FileNotFoundError(f"{direc} does not contain E-field data")

    time = datetime_range(cfg["time"][0], cfg["time"][0] + cfg["tdur"], cfg["dtE0"])
    for t in time:
        file = find.frame(path, t)
        if not file:
            logging.error(f"no E-field data found at {t} in {path}")
            continue

        dat = read.Efield(file)

        for k in {"Exit", "Eyit", "Vminx1it", "Vmaxx1it", "Vminx2ist", "Vmaxx2ist"}:
            if dat[k].ndim == 1:
                fg = plot2d_input(dat[k], cfg)
            else:
                fg = plot3d_input(dat[k], cfg)

            fg.suptitle(f"{k}: {t}")

            save_fig(fg, direc, f"Efield-{k}", time=t)


def precip(direc: Path):
    """plot input precipitation

    Parameters
    ----------

    direc: pathlib.Path
        top-level simulation directory
    """

    direc = Path(direc).expanduser()

    cfg = read.config(direc)
    precip_path = find.inputs(direc, cfg.get("precdir"))
    if not precip_path:
        raise FileNotFoundError(f"{direc} does not contain precipitation data")

    time = datetime_range(cfg["time"][0], cfg["time"][0] + cfg["tdur"], cfg["dtprec"])

    for t in time:
        file = find.frame(precip_path, t)
        if not file:
            logging.error(f"no precipitation data found at {t} in {precip_path}")
            continue

        dat = read.precip(file)

        for k in {"E0", "Q"}:
            if dat[k].ndim == 1:
                fg = plot2d_input(dat[k], cfg)
            else:
                fg = plot3d_input(dat[k], cfg)

            fg.suptitle(f"{k}: {t}")

            save_fig(fg, direc, f"precip-{k}", time=t)


def plot2d_input(A, cfg: dict[str, T.Any]) -> Figure:
    fg = Figure()
    ax = fg.gca()

    if cfg["lyp"] == 1:
        x = A["mlon"]
        ax.set_xlabel("magnetic longitude")
    else:
        x = A["mlat"]
        ax.set_xlabel("magnetic latitude")

    ax.plot(x, A)

    return fg


def plot3d_input(A, cfg: dict[str, T.Any]) -> Figure:
    fg = Figure()
    ax = fg.gca()

    h0 = ax.pcolormesh(A["mlon"], A["mlat"], A, shading="nearest")
    fg.colorbar(h0, ax=ax)
    ax.set_ylabel("magnetic latitude")
    ax.set_xlabel("magnetic longitude")

    return fg


def basic(xg: dict[str, T.Any]) -> Figure:
    fig = Figure()
    axs = fig.subplots(1, 3)
    # %% x1
    lx1 = xg["x1"].size
    ax = axs[0]
    ax.plot(range(lx1), xg["x1"] / 1000, marker=".")
    ax.set_ylabel("x1 [km]")
    ax.set_xlabel("index (dimensionless)")
    ax.set_title(f"x1 (upward) lx1 = {lx1}")

    # %% x2
    lx2 = xg["x2"].size
    ax = axs[1]
    ax.plot(xg["x2"] / 1e3, range(lx2), marker=".")
    ax.set_xlabel("x2 [km]")
    ax.set_ylabel("index (dimensionless)")
    ax.set_title(f"x2 (eastward) lx2 = {lx2}")

    # %% x3
    lx3 = xg["x3"].size
    ax = axs[2]
    ax.plot(range(lx3), xg["x3"] / 1e3, marker=".")
    ax.set_ylabel("x3 [km]")
    ax.set_xlabel("index (dimensionless)")
    ax.set_title(f"x3 (northward) lx3 = {lx3}")

    return fig


def stitle(fig: Figure, xg: dict[str, T.Any], ttxt: str = ""):
    """suptitle"""
    if "time" in xg:
        ttxt += f" {xg['time']}"

    if "filename" in xg:
        ttxt += f" {xg['filename']}"

    fig.suptitle(ttxt)


def altitude_grid(xg: dict[str, T.Any]) -> Figure:
    """
    plot altitude x1 grid

    Parameters
    ----------

    xg: pathlib.Path or dict of numpy.ndarray
        simulation grid: filename or dict
    """

    if isinstance(xg, Path):
        xg = read.grid(xg)

    x1_km = xg["x1"] / 1000

    fig = Figure()
    ax = fig.gca()

    ax.plot(x1_km, marker="*")
    ax.set_ylabel("x1 [km]")
    ax.set_xlabel("index (dimensionless)")

    file = xg.get("filename", "")

    ax.set_title(
        f"{file}  min. alt: {x1_km.min():0.1f} [km]  max. alt: {x1_km.max():0.1f} [km]  lx1: {x1_km.size}"
    )

    return fig


def plot_3d(direc: Path, var: list[str], saveplot_fmt: str = None):
    from . import vis3d

    direc = Path(direc).expanduser().resolve(strict=True)

    cfg = read.config(direc)
    for t in cfg["time"]:
        vis3d.frame(direc, time=t, var=var, saveplot_fmt=saveplot_fmt)


def plot_all(direc: Path, var: list[str] = None, saveplot_fmt: str = None):

    direc = Path(direc).expanduser().resolve(strict=True)

    if not var:
        var = PARAMS
    if isinstance(var, str):
        var = [var]

    if set(var).intersection({"png", "pdf", "eps"}):
        raise ValueError("please use saveplot_fmt='png' or similar for plot format")

    cfg = read.config(direc)
    # %% loop over files / time
    for t in cfg["time"]:
        frame(direc, time=t, var=var, saveplot_fmt=saveplot_fmt)


def frame(
    direc: Path,
    time: datetime = None,
    *,
    saveplot_fmt: str = None,
    var: list[str] = None,
    xg: dict[str, T.Any] = None,
):
    """
    if save_dir, plots will not be visible while generating to speed plot writing
    """

    if not var:
        var = PARAMS

    file = None
    if time is None:
        if not direc.is_file():
            raise ValueError("must either specify directory and time, or single file")
        file = direc
        direc = direc.parent

    if not xg:
        xg = read.grid(direc)

    if file is None:
        dat = read.frame(direc, time, var=var)
    else:
        dat = read.data(file, var)

    if not dat:
        raise ValueError(f"No data in {direc} at {time}")

    plotfun = grid2plotfun(xg)

    for k, v in dat.items():
        if any(s in k for s in var):
            fg = plotfun(dat.time, xg, v.squeeze(), k, wavelength=dat.get("wavelength"))
            save_fig(fg, direc, k, saveplot_fmt, time)


def save_fig(fg: Figure, direc: Path, name: str, fmt: str = "png", time: datetime = None):
    if not fmt:
        fmt = "png"

    if time is None:
        tstr = ""
    else:
        tstr = f"-{time.isoformat().replace(':','')}"

    plot_fn = direc / f"plots/{name}{tstr}.{fmt}"
    plot_fn.parent.mkdir(exist_ok=True)
    print(f"{time} => {plot_fn}")
    fg.savefig(plot_fn)



"""
3-D visualizations using Mayavi / VTK

suggested install:

    conda install mayavi
"""

from __future__ import annotations
from pathlib import Path
import numpy as np
from datetime import datetime
import typing as T

# import scipy.interpolate as interp

from mayavi import mlab

# mlab.options.backend = 'envisage'  # for GUI

from . import PARAMS
from .. import read

PLOTFUN = {"scalar": ("ne", "Ti", "Te", "J1", "J2", "J3"), "vector": ("v1", "v2", "v3")}
R_EARTH = 6370e3


def frame(
    direc: Path,
    time: datetime,
    var: list[str] = None,
    saveplot_fmt: str = None,
    xg: dict[str, T.Any] = None,
):
    """
    plot plasma quantities in 3D

    if save_dir, plots will not be visible while generating to speed plot writing
    """
    if not var:
        var = PARAMS

    if not xg:
        xg = read.grid(direc)

    dat = read.frame(direc, time)
    time = dat["time"]

    for k in var:
        if k not in dat:  # not present at this time step, often just the first time step
            continue

        if k in PLOTFUN["scalar"]:
            fg = scalar(time, xg, dat[k].squeeze(), name=k)
        elif k in PLOTFUN["vector"]:
            print("TODO: vector plot", k)

        if saveplot_fmt:
            plot_fn = direc / "plots" / f"{k}-{time.isoformat().replace(':','')}.{saveplot_fmt}"
            plot_fn.parent.mkdir(exist_ok=True)
            print(f"{dat['time']} => {plot_fn}")
            fg.savefig(plot_fn)


def scalar(time: datetime, xg: dict[str, np.ndarray], parm: np.ndarray, name: str):
    """
    plot scalar field data in transparent 3D volume
    """

    if parm.ndim != 3:
        raise ValueError("Mayavi scalar field plots are for 3-D data")

    # %% arbitrary output plot resolution
    # lxp = 100
    # lyp = 100
    # lzp = 100

    # %% SIZE OF SIMULATION
    lxs = read.get_lxs(xg)

    lx2 = lxs[1]
    # inds1 = slice(2, lx1 + 2)
    inds2 = slice(2, lx2 + 2)
    # inds3 = slice(2, lx3 + 2)

    # %% SIZE OF PLOT GRID THAT WE ARE INTERPOLATING ONTO
    meantheta = xg["theta"].mean()
    # this is a mag colat. coordinate and is only used for defining grid in linspaces below
    # runs backward from north distance, hence the negative sign

    # northward distance [m]
    y = -(xg["theta"] - meantheta) * R_EARTH
    # eastward distance [m]
    x = xg["x2"][inds2]
    # upward distance [m]
    z = xg["alt"]

    # Mayavi requires a grid like so:
    # interpolatedz
    # eastward distance [meters]
    # xp = np.linspace(x.min(), x.max(), lxp)
    # northward distance [meters]
    # yp = np.linspace(y.min(), y.max(), lyp)
    # upward distance [meters]
    # zp = np.linspace(z.min(), z.max(), lzp)
    # x3, y3, z3 = np.mgrid[xp[0]: xp[-1]: lxp * 1j,
    #   yp[0]: yp[-1]: lyp * 1j, zp[0]: zp[-1]: lzp * 1j]  # type: ignore

    # non-interpolated
    parm = parm.transpose(1, 2, 0)
    xp = np.linspace(x.min(), x.max(), parm.shape[0])
    yp = np.linspace(y.min(), y.max(), parm.shape[1])
    zp = np.linspace(z.min(), z.max(), parm.shape[2])
    x3, y3, z3 = np.mgrid[
        xp[0] : xp[-1] : xp.size * 1j, yp[0] : yp[-1] : yp.size * 1j, zp[0] : zp[-1] : zp.size * 1j  # type: ignore
    ]

    # %% 3-D interpolation for plot
    if ~np.isfinite(parm).all():
        raise ValueError("Mayavi requires finite data values")

    # TODO: check order of interpolated axes (1,2,0) or ?
    # parmp = interp.interpn(
    #     points=(xg["x1"][inds1], xg["x2"][inds2], xg["x3"][inds3]),
    #     values=parm.values,
    #     xi=np.column_stack((x3.ravel(), y3.ravel(), z3.ravel())),
    #     bounds_error=False,
    # ).reshape((lxp, lyp, lzp))

    # if ~np.isfinite(parmp).all():
    #     raise ValueError('Interpolation issue: Mayavi requires finite data values')

    fig = mlab.figure()
    scf = mlab.pipeline.scalar_field(x3 / 1e3, y3 / 1e3, z3 / 1e3, parm, figure=fig)
    vol = mlab.pipeline.volume(scf, figure=fig)
    mlab.colorbar(vol, title=name)
    mlab.axes(figure=fig, xlabel="x (km)", ylabel="y (km)", zlabel="z (km)")

    return fig



from __future__ import annotations
import typing as T
from pathlib import Path
import subprocess
import shutil
import json
import os
import tempfile
import importlib.resources

from .. import PYGEMINI_ROOT
from ..utils import get_cpu_count
from ..web import git_download

__all__ = ["exe", "build", "find_library"]


def exe() -> str:

    cmake = shutil.which("cmake")
    if not cmake:
        raise FileNotFoundError("CMake not found.")

    cmake_version = (
        subprocess.check_output([cmake, "--version"], text=True).split("\n")[0].split(" ")[2]
    )

    print("Using CMake", cmake_version)

    return cmake


def build(
    source_dir: Path,
    build_dir: Path,
    *,
    config_args: list[str] = None,
    build_args: list[str] = None,
    wipe: bool = False,
    env: T.Mapping[str, str] = None,
    run_test: bool = False,
    dryrun: bool = False,
    install: bool = True,
):
    """ build and install with CMake """
    cmake = exe()

    cache_file = build_dir / "CMakeCache.txt"
    if wipe:
        if cache_file.is_file():
            cache_file.unlink()
    # %% Configure
    cmd = [cmake, f"-B{build_dir}", f"-S{source_dir}"]
    if config_args:
        cmd += config_args
    subprocess.check_call(cmd, env=env)
    # %% Build
    cmd = [cmake, "--build", str(build_dir), "--parallel"]
    if build_args:
        cmd += build_args
    if dryrun:
        print("DRYRUN: would have run\n", " ".join(cmd))
        return None

    subprocess.check_call(cmd)

    if run_test:
        Njobs = get_cpu_count()
        subprocess.check_call(
            ["ctest", "--parallel", str(Njobs), "--output-on-failure"], cwd=str(build_dir)
        )

    if install:
        subprocess.check_call([cmake, "--install", str(build_dir)])


def find_library(lib_name: str, lib_path: list[str], env: T.Mapping[str, str]) -> bool:
    """
    check if library exists with CMake

    lib_name must have the appropriate upper and lower case letter as would be used
    directly in CMake.
    """

    cmake = exe()

    with importlib.resources.path("gemini3d.cmake", "FindLAPACK.cmake") as f:
        mod_path = Path(f).parent

    cmake_template = """
cmake_minimum_required(VERSION 3.15)
project(dummy LANGUAGES C Fortran)

"""

    if mod_path.is_dir():
        cmake_template += f'list(APPEND CMAKE_MODULE_PATH "{mod_path.as_posix()}")\n'

    cmake_template += f"find_package({lib_name} REQUIRED)\n"

    build_dir = f"find-{lib_name.split(' ', 1)[0]}"

    # not context_manager to avoid Windows PermissionError on context exit for Git subdirs
    d = tempfile.TemporaryDirectory()
    r = Path(d.name)
    (r / "CMakeLists.txt").write_text(cmake_template)

    cmd = [cmake, "-S", str(r), "-B", str(r / build_dir)] + lib_path
    # use cwd= to avoid spilling temporary files into current directory if ancient CMake used
    # also avoids bugs if there is a CMakeLists.txt in the current directory
    ret = subprocess.run(cmd, env=env, cwd=str(r))

    try:
        d.cleanup()
    except PermissionError:
        pass

    return ret.returncode == 0


def build_gemini3d(target: Path) -> Path:
    """
    build targets from gemini3d program

    Specify environemnt variable GEMINI_ROOT to reuse existing development code
    """
    src_dir = None
    if os.environ.get("GEMINI_ROOT"):
        src_dir = Path(os.environ["GEMINI_ROOT"]).expanduser()

    if not src_dir or not src_dir.is_dir():
        if (
            len(target.parents) >= 2
            and target.parents[1].is_dir()
            and (target.parents[1] / "CMakeLists.txt").is_file()
        ):
            src_dir = target.parents[1]
        else:
            # clone Gemini3D and do a test build

            src_dir = PYGEMINI_ROOT / "gemini-fortran"

            jmeta = json.loads(importlib.resources.read_text("gemini3d", "libraries.json"))
            git_download(src_dir, repo=jmeta["gemini3d"]["git"], tag=jmeta["gemini3d"]["tag"])

    if not src_dir.is_dir():
        raise NotADirectoryError(f"could not find Gemini3D source directory {src_dir}")

    build_dir = src_dir / "build"
    exe = shutil.which(target.name, path=str(build_dir))

    if not exe:
        build_args = ["--target", target.name]

        build(
            src_dir,
            build_dir,
            run_test=False,
            install=False,
            config_args=["-DBUILD_TESTING:BOOL=false"],
            build_args=build_args,
        )
        exe = shutil.which(target.name, path=str(build_dir))

    if not exe:
        raise RuntimeError(f"{target.name} not found in {build_dir}")

    return Path(exe)



"""
create simulation grids
"""

from __future__ import annotations
import logging
import typing as T
import math
import numpy as np

from . import read
from .coord import geog2geomag, geomag2geog


def cart3d(p: dict[str, T.Any]) -> dict[str, T.Any]:
    """make 3D grid and save to disk

    Parameters
    -----------

    p: dict
        simulation parameters

    Returns
    -------

    xg: dict
        simulation grid
    """

    # %%create altitude grid
    # original Matlab params
    # p.alt_min = 80e3;
    # p.alt_max = 1000e3;
    # p.alt_scale = [10e3, 8e3, 500e3, 150e3];

    if {"alt_min", "alt_max", "alt_scale", "Bincl"} <= p.keys():
        # https://docs.python.org/3/library/stdtypes.html#frozenset.issubset
        z = altitude_grid(p["alt_min"], p["alt_max"], p["Bincl"], p["alt_scale"])
    elif "eq_dir" in p and p["eq_dir"].is_file():
        logging.info(f"makegrid_cart_3D: reusing grid from {p['eq_dir']}")
        xeq = read.grid(p["eq_dir"])
        z = xeq["x1"]
        del xeq
    else:
        raise ValueError("must specify altitude grid parameters or grid file to reuse")

    # %% TRANSVERSE GRID (BASED ON SIZE OF CURRENT REGION SPECIFIED ABOVE)
    # EAST
    if "x2parms" in p:
        x = grid1d(p["xdist"], p["lxp"], p["x2parms"])
    else:
        x = grid1d(p["xdist"], p["lxp"])

    # NORTH
    if "x3parms" in p:
        y = grid1d(p["ydist"], p["lyp"], p["x3parms"])
    else:
        y = grid1d(p["ydist"], p["lyp"])

    # %% COMPUTE CELL WALL LOCATIONS
    lx2 = x.size
    xi = np.empty(lx2 + 1)
    xi[1:-1] = 1 / 2 * (x[1:] + x[:-1])
    xi[0] = x[0] - 1 / 2 * (x[1] - x[0])
    xi[-1] = x[-1] + 1 / 2 * (x[-1] - x[-2])

    lx3 = y.size
    yi = np.empty(lx3 + 1)
    yi[1:-1] = 1 / 2 * (y[1:] + y[:-1])
    yi[0] = y[0] - 1 / 2 * (y[1] - y[0])
    yi[-1] = y[-1] + 1 / 2 * (y[-1] - y[-2])

    lx1 = z.size
    zi = np.empty(lx1 + 1)
    zi[1:-1] = 1 / 2 * (z[1:] + z[:-1])
    zi[0] = z[0] - 1 / 2 * (z[1] - z[0])
    zi[-1] = z[-1] + 1 / 2 * (z[-1] - z[-2])

    # %% GRAVITATIONAL FIELD COMPONENTS IN DIPOLE SYSTEM
    Re = 6370e3
    G = 6.67428e-11
    Me = 5.9722e24
    r = z + Re
    g = G * Me / r ** 2
    gz = np.broadcast_to(-g[:, None, None], (g.size, lx2, lx3))
    assert gz.shape == (lx1, lx2, lx3)

    # DISTANCE EW AND NS (FROM ENU (or UEN in our case - cyclic permuted) COORD. SYSTEM)
    # #NEED TO BE CONVERTED TO DIPOLE SPHERICAL AND THEN
    # GLAT/GLONG - BASICALLY HERE WE ARE MAPPING THE CARTESIAN GRID ONTO THE
    # SURFACE OF A SPHERE THEN CONVERTING TO GEOGRAPHIC.
    # get the magnetic coordinates of the grid center, based on user input
    thetactr, phictr = geog2geomag(p["glat"], p["glon"])

    # %% Center of earth distance
    r = Re + z
    r = np.broadcast_to(r[:, None, None], (r.size, lx2, lx3))
    assert r.shape == (lx1, lx2, lx3)

    # %% Northward angular distance
    gamma2 = y / Re
    # must retain the sign of x3
    theta = thetactr - gamma2
    # minus because distance north is against theta's direction
    theta = np.broadcast_to(theta[None, None, :], (lx1, lx2, theta.size))
    assert theta.shape == (lx1, lx2, lx3)

    # %% Eastward angular distance
    # gamma1=x/Re;     %must retain the sign of x2
    gamma1 = x / Re / np.sin(thetactr)
    # must retain the sign of x2, just use theta of center of grid
    phi = phictr + gamma1
    phi = np.broadcast_to(phi[None, :, None], (lx1, phi.size, lx3))
    assert phi.shape == (lx1, lx2, lx3)

    # %% COMPUTE THE GEOGRAPHIC COORDINATES OF EACH GRID POINT
    glatgrid, glongrid = geomag2geog(theta, phi)

    # %% COMPUTE ECEF CARTESIAN IN CASE THEY ARE NEEDED
    xECEF = r * np.sin(theta) * np.cos(phi)
    yECEF = r * np.sin(theta) * np.sin(phi)
    zECEF = r * np.cos(theta)

    # %% COMPUTE SPHERICAL ECEF UNIT VECTORS - CARTESIAN-ECEF COMPONENTS
    er = np.empty((lx1, lx2, lx3, 3))
    etheta = np.empty_like(er)
    ephi = np.empty_like(er)

    er[:, :, :, 0] = np.sin(theta) * np.cos(phi)
    # xECEF-component of er
    er[:, :, :, 1] = np.sin(theta) * np.sin(phi)
    # yECEF
    er[:, :, :, 2] = np.cos(theta)
    # zECEF
    etheta[:, :, :, 0] = np.cos(theta) * np.cos(phi)
    etheta[:, :, :, 1] = np.cos(theta) * np.sin(phi)
    etheta[:, :, :, 2] = -np.sin(theta)
    ephi[:, :, :, 0] = -np.sin(phi)
    ephi[:, :, :, 1] = np.cos(phi)
    ephi[:, :, :, 2] = 0

    # %% UEN UNIT VECTORS IN ECEF COMPONENTS
    e1 = er
    # up is the same direction as from ctr of earth
    e2 = ephi
    # e2 is same as ephi
    e3 = -etheta
    # etheta is positive south, e3 is pos. north

    # %% STORE RESULTS IN GRID DATA STRUCTURE
    xg = {
        "x1": z,
        "x2": x,
        "x3": y,
        "x1i": zi,
        "x2i": xi,
        "x3i": yi,
    }

    lx = (xg["x1"].size, xg["x2"].size, xg["x3"].size)
    xg["lx"] = np.array(lx)

    xg["dx1f"] = np.append(xg["x1"][1:] - xg["x1"][:-1], xg["x1"][-1] - xg["x1"][-2])
    # FWD DIFF
    xg["dx1b"] = np.insert(xg["x1"][1:] - xg["x1"][:-1], 0, xg["x1"][1] - xg["x1"][0])
    # BACK DIFF
    xg["dx1h"] = xg["x1i"][1:-1] - xg["x1i"][:-2]
    # MIDPOINT DIFFS

    xg["dx2f"] = np.append(xg["x2"][1:] - xg["x2"][:-1], xg["x2"][-1] - xg["x2"][-2])
    # FWD DIFF
    xg["dx2b"] = np.insert(xg["x2"][1:] - xg["x2"][:-1], 0, xg["x2"][1] - xg["x2"][0])
    # BACK DIFF
    xg["dx2h"] = xg["x2i"][1:-1] - xg["x2i"][:-2]
    # MIDPOINT DIFFS

    xg["dx3f"] = np.append(xg["x3"][1:] - xg["x3"][:-1], xg["x3"][-1] - xg["x3"][-2])
    # FWD DIFF
    xg["dx3b"] = np.insert(xg["x3"][1:] - xg["x3"][:-1], 0, xg["x3"][1] - xg["x3"][0])
    # BACK DIFF
    xg["dx3h"] = xg["x3i"][1:-1] - xg["x3i"][:-2]
    # MIDPOINT DIFFS

    xg["h1"] = np.ones(lx)
    xg["h2"] = np.ones(lx)
    xg["h3"] = np.ones(lx)
    xg["h1x1i"] = np.ones((lx[0] + 1, lx[1], lx[2]))
    xg["h2x1i"] = np.ones((lx[0] + 1, lx[1], lx[2]))
    xg["h3x1i"] = np.ones((lx[0] + 1, lx[1], lx[2]))
    xg["h1x2i"] = np.ones((lx[0], lx[1] + 1, lx[2]))
    xg["h2x2i"] = np.ones((lx[0], lx[1] + 1, lx[2]))
    xg["h3x2i"] = np.ones((lx[0], lx[1] + 1, lx[2]))
    xg["h1x3i"] = np.ones((lx[0], lx[1], lx[2] + 1))
    xg["h2x3i"] = np.ones((lx[0], lx[1], lx[2] + 1))
    xg["h3x3i"] = np.ones((lx[0], lx[1], lx[2] + 1))

    # %% Cartesian, ECEF representation of curvilinar coordinates
    xg["e1"] = e1
    xg["e2"] = e2
    xg["e3"] = e3

    # %% ECEF spherical coordinates
    xg["r"] = r
    xg["theta"] = theta
    xg["phi"] = phi
    # xg.rx1i=[]; xg.thetax1i=[];
    # xg.rx2i=[]; xg.thetax2i=[];

    # %% These are cartesian representations of the ECEF, spherical unit vectors
    xg["er"] = er
    xg["etheta"] = etheta
    xg["ephi"] = ephi

    xg["I"] = np.broadcast_to(p["Bincl"], (lx2, lx3))

    # %% Cartesian ECEF coordinates
    xg["x"] = xECEF
    xg["z"] = zECEF
    xg["y"] = yECEF
    xg["alt"] = xg["r"] - Re
    # since we need a 3D array use xg.r here...

    xg["gx1"] = gz
    xg["gx2"] = np.zeros(lx)
    xg["gx3"] = np.zeros(lx)

    xg["Bmag"] = np.broadcast_to(-50000e-9, xg["lx"])
    # minus for northern hemisphere...

    xg["glat"] = glatgrid
    xg["glon"] = glongrid

    # xg['xp']=x; xg['zp']=z;

    # xg['inull']=[];
    xg["nullpts"] = np.zeros(lx)

    # %% TRIM DATA STRUCTRE TO BE THE SIZE FORTRAN EXPECTS
    # note: xgf is xg == True
    xgf = xg

    # indices corresponding to non-ghost cells for 1 dimension
    i1 = slice(2, lx[0] - 2)
    i2 = slice(2, lx[1] - 2)
    i3 = slice(2, lx[2] - 2)

    # any dx variable will not need to first element (backward diff of two ghost cells)
    idx1 = slice(1, lx[0])
    idx2 = slice(1, lx[1])
    idx3 = slice(1, lx[2])

    # x1-interface variables need only non-ghost cell values (left interface) plus one
    ix1i = slice(2, lx[0] - 1)
    ix2i = slice(2, lx[1] - 1)
    ix3i = slice(2, lx[2] - 1)

    # remove ghost cells
    # now that indices have been define we can go ahead and make this change
    xgf["lx"] = xgf["lx"] - 4

    xgf["dx1b"] = xgf["dx1b"][idx1]
    xgf["dx2b"] = xgf["dx2b"][idx2]
    xgf["dx3b"] = xgf["dx3b"][idx3]

    xgf["x1i"] = xgf["x1i"][ix1i]
    xgf["x2i"] = xgf["x2i"][ix2i]
    xgf["x3i"] = xgf["x3i"][ix3i]

    xgf["dx1h"] = xgf["dx1h"][i1]
    xgf["dx2h"] = xgf["dx2h"][i2]
    xgf["dx3h"] = xgf["dx3h"][i3]

    xgf["h1x1i"] = xgf["h1x1i"][ix1i, i2, i3]
    xgf["h2x1i"] = xgf["h2x1i"][ix1i, i2, i3]
    xgf["h3x1i"] = xgf["h3x1i"][ix1i, i2, i3]

    xgf["h1x2i"] = xgf["h1x2i"][i1, ix2i, i3]
    xgf["h2x2i"] = xgf["h2x2i"][i1, ix2i, i3]
    xgf["h3x2i"] = xgf["h3x2i"][i1, ix2i, i3]

    xgf["h1x3i"] = xgf["h1x3i"][i1, i2, ix3i]
    xgf["h2x3i"] = xgf["h2x3i"][i1, i2, ix3i]
    xgf["h3x3i"] = xgf["h3x3i"][i1, i2, ix3i]

    xgf["gx1"] = xgf["gx1"][i1, i2, i3]
    xgf["gx2"] = xgf["gx2"][i1, i2, i3]
    xgf["gx3"] = xgf["gx3"][i1, i2, i3]

    xgf["glat"] = xgf["glat"][i1, i2, i3]
    xgf["glon"] = xgf["glon"][i1, i2, i3]
    xgf["alt"] = xgf["alt"][i1, i2, i3]

    xgf["Bmag"] = xgf["Bmag"][i1, i2, i3]

    xgf["I"] = xgf["I"][i2, i3]

    xgf["nullpts"] = xgf["nullpts"][i1, i2, i3]

    xgf["e1"] = xgf["e1"][i1, i2, i3, :]
    xgf["e2"] = xgf["e2"][i1, i2, i3, :]
    xgf["e3"] = xgf["e3"][i1, i2, i3, :]

    xgf["er"] = xgf["er"][i1, i2, i3, :]
    xgf["etheta"] = xgf["etheta"][i1, i2, i3, :]
    xgf["ephi"] = xgf["ephi"][i1, i2, i3, :]

    xgf["r"] = xgf["r"][i1, i2, i3]
    xgf["theta"] = xgf["theta"][i1, i2, i3]
    xgf["phi"] = xgf["phi"][i1, i2, i3]

    xgf["x"] = xgf["x"][i1, i2, i3]
    xgf["y"] = xgf["y"][i1, i2, i3]
    xgf["z"] = xgf["z"][i1, i2, i3]

    return xgf


def grid1d(dist: float, L: int, parms: list[float] = None) -> np.ndarray:
    """
    generate 1D grid

    Parameters
    ----------

    dist: float
        one-way distance from origin (meters)
    L: int
        number of cells

    Returns
    -------

    x: np.ndarray
        1D vector grid
    """

    xmin = -dist / 2
    xmax = dist / 2

    if not parms:
        x = uniform1d(xmin, xmax, L)
    else:
        x = non_uniform1d(xmax, parms)

    return x


def uniform1d(xmin: float, xmax: float, L: int) -> np.ndarray:

    if L == 1:
        # degenerate dimension
        # add 2 ghost cells on each side
        x = np.linspace(xmin, xmax, L + 4)
    else:
        # exclude the ghost cells when setting extents
        x = np.linspace(xmin, xmax, L)
        d0 = x[1] - x[0]
        d1 = x[-1] - x[-2]
        # now tack on ghost cells so they are outside user-specified region
        x = np.insert(x, 0, [x[0] - 2 * d0, x[0] - d0])
        x = np.append(x, [x[-1] + d1, x[-1] + 2 * d1])

    return x


def non_uniform1d(xmax: float, parms: list[float]) -> np.ndarray:

    degdist = parms[0]  # distance from boundary at which we start to degrade resolution
    dx0 = parms[1]  # min step size for grid
    dxincr = parms[2]  # max step size increase for grid
    ell = parms[3]  # transition length of degradation
    x2 = xmax - degdist

    x = [dx0 / 2.0]
    # start offset from zero so we can have an even number (better for mpi)

    while x[-1] < xmax:
        dx = dx0 + dxincr * (1 / 2 + 1 / 2 * math.tanh((x[-1] - x2) / ell))
        x.append(x[-1] + dx)

    x = np.append(-np.array(x[::-1]), x)

    return x


def altitude_grid(
    alt_min: float, alt_max: float, incl_deg: float, d: tuple[float, float, float, float]
) -> np.ndarray:

    if alt_min < 0 or alt_max < 0:
        raise ValueError("grid values must be positive")
    if alt_max <= alt_min:
        raise ValueError("grid_max must be greater than grid_min")

    alt = [alt_min]

    while alt[-1] < alt_max:
        # dalt=10+9.5*tanh((alt(i-1)-500)/150)
        dalt = d[0] + d[1] * math.tanh((alt[-1] - d[2]) / d[3])
        alt.append(alt[-1] + dalt)

    alt = np.asarray(alt)
    if alt.size < 10:
        raise ValueError("grid too small")

    # %% tilt for magnetic inclination
    z = alt / math.sin(math.radians(incl_deg))

    # %% add two ghost cells each to top and bottom
    dz1 = z[1] - z[0]
    dzn = z[-1] - z[-2]
    z = np.insert(z, 0, [z[0] - 2 * dz1, z[0] - dz1])
    z = np.append(z, [z[-1] + dzn, z[-1] + 2 * dzn])

    return z
